"""Utility functions to work with the Zoltan generated lists"""
import numpy
from mpi4py import MPI

def get_num_objects_per_proc(comm, num_objects):
    """Utility function to get number of objects on each processor.

    Parameters:
    -----------

    comm : mpi.Comm
        The communicator (COMM_WORLD)

    num_objects : int
        Number of objects per processor

    This function uses MPI.Allreduce to get, on each processor, an
    array of size 'comm.Get_size()' which stores the number of
    particles per processor. Using this array, we can compute the new
    unique global indices.

    """
    size = comm.Get_size()
    rank = comm.Get_rank()

    send_data = numpy.zeros(shape=size, dtype=numpy.int32)
    send_data[rank] = num_objects

    num_objects_data = numpy.zeros(shape=size, dtype=numpy.int32)
    comm.Allreduce(send_data, num_objects_data, op=MPI.MAX)

    return num_objects_data



"""3 dimensional tests for the Zoltan partitioner. The test follows
the same pattern as the 2D test.

To see the output from the script try the following::

    $ mpirun -np 4 python 3d_partition.py --plot

"""
import sys

import mpi4py.MPI as mpi
comm = mpi.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

from cyarray.api import UIntArray, DoubleArray
from pyzoltan.core import zoltan

from numpy import random
import numpy as np


colors = ['r', 'g', 'b', 'y', 'm', 'k', 'c', 'burlywood']

def plot_points(x, y, z, slice_data, title, filename):
    if '--plot' not in sys.argv:
        return

    from mpl_toolkits.mplot3d import Axes3D
    import matplotlib.pyplot as plt
    fig = plt.figure()
    s1 = fig.add_subplot(111)
    s1.axes = Axes3D(fig)
    for i in range(size):
        s1.axes.plot3D(
            x[slice_data[i]], y[slice_data[i]], z[slice_data[i]],
            c=colors[i], marker='o', linestyle='None', alpha=0.5
        )

    s1.axes.set_xlabel( 'X' )
    s1.axes.set_ylabel( 'Y' )
    s1.axes.set_zlabel( 'Z' )

    plt.title(title)
    plt.savefig(filename)


numPoints = 1<<12

x = random.random( numPoints )
y = random.random( numPoints )
z = random.random( numPoints )
gid = np.arange( numPoints*size, dtype=np.uint32 )[rank*numPoints:(rank+1)*numPoints]

X = np.zeros( size * numPoints )
Y = np.zeros( size * numPoints )
Z = np.zeros( size * numPoints )
GID = np.arange( numPoints*size, dtype=np.uint32 )

comm.Gather( sendbuf=x, recvbuf=X, root=0 )
comm.Gather( sendbuf=y, recvbuf=Y, root=0 )
comm.Gather( sendbuf=z, recvbuf=Z, root=0 )

if rank == 0:
    slice_data = [slice(i*numPoints, (i+1)*numPoints) for i in range(size)]
    plot_points(
        X, Y, Z, slice_data, title="Initial Distribution",
        filename="initial.pdf"
    )
# partition the points using PyZoltan
xa = DoubleArray(numPoints); xa.set_data(x)
ya = DoubleArray(numPoints); ya.set_data(y)
za = DoubleArray(numPoints); za.set_data(z)
gida = UIntArray(numPoints); gida.set_data(gid)

# create the geometric partitioner
pz = zoltan.ZoltanGeometricPartitioner(
    dim=3, comm=comm, x=xa, y=ya, z=za, gid=gida)

# call the load balancing function
pz.set_lb_method('RIB')
pz.Zoltan_Set_Param('DEBUG_LEVEL', '1')
pz.Zoltan_LB_Balance()

# get the new assignments
my_global_ids = list( gid )

# remove points to be exported
for i in range(pz.numExport):
    my_global_ids.remove( pz.exportGlobalids[i] )

# add points to be imported
for i in range(pz.numImport):
    my_global_ids.append( pz.importGlobalids[i] )

new_gids = np.array( my_global_ids, dtype=np.uint32 )

# gather the new gids on root as a list
NEW_GIDS = comm.gather( new_gids, root=0 )

# save the new partition
if rank == 0:
    plot_points(
        X, Y, Z, NEW_GIDS,
        title='Final Distribution', filename='final.pdf'
    )
comm.barrier()



# cython: language_level=3, embedsignature=True
# distutils: language=c++

import numpy as np
cimport numpy as np

# Zoltan comm objects and headers
from pyzoltan.czoltan cimport czoltan_comm as zcomm

cdef class ZComm:
    # pointer to the zoltan comm object
    cdef zcomm.ZOLTAN_COMM_OBJ* _zoltan_comm_obj

    # mpi communicator, rank and size
    cdef public object comm
    cdef public int rank, size

    # tag for the communicator
    cdef public int tag

    # number of objects to send
    cdef public int nsend

    # processor list for particles to be exported
    cdef public np.ndarray proclist

    # number of objects to be received
    cdef public int nreturn

    # size of each element and dtype
    cdef public int nbytes
    cdef public object dtype



# cython: language_level=3, embedsignature=True
# distutils: language=c++

"""Cython Wrapper for Zoltan. """

if MPI4PY_V2:
   from mpi4py.libmpi cimport MPI_Comm
else:
   from mpi4py.mpi_c cimport MPI_Comm

from czoltan_types cimport ZOLTAN_ID_PTR

cdef extern from "zoltan.h":

    # Zoltan version number
    float ZOLTAN_VERSION_NUMBER

    # ****************************************************************** #
    # *  Enumerated type used to indicate which function is to be set by
    # *  ZOLTAN_Set_Fn.
    enum Zoltan_Fn_Type:
        ZOLTAN_NUM_EDGES_FN_TYPE
        ZOLTAN_NUM_EDGES_MULTI_FN_TYPE
        ZOLTAN_EDGE_LIST_FN_TYPE
        ZOLTAN_EDGE_LIST_MULTI_FN_TYPE
        ZOLTAN_NUM_GEOM_FN_TYPE
        ZOLTAN_GEOM_MULTI_FN_TYPE
        ZOLTAN_GEOM_FN_TYPE
        ZOLTAN_NUM_OBJ_FN_TYPE
        ZOLTAN_OBJ_LIST_FN_TYPE
        ZOLTAN_FIRST_OBJ_FN_TYPE
        ZOLTAN_NEXT_OBJ_FN_TYPE
        ZOLTAN_NUM_BORDER_OBJ_FN_TYPE
        ZOLTAN_BORDER_OBJ_LIST_FN_TYPE
        ZOLTAN_FIRST_BORDER_OBJ_FN_TYPE
        ZOLTAN_NEXT_BORDER_OBJ_FN_TYPE
        ZOLTAN_PRE_MIGRATE_PP_FN_TYPE
        ZOLTAN_MID_MIGRATE_PP_FN_TYPE
        ZOLTAN_POST_MIGRATE_PP_FN_TYPE
        ZOLTAN_PRE_MIGRATE_FN_TYPE
        ZOLTAN_MID_MIGRATE_FN_TYPE
        ZOLTAN_POST_MIGRATE_FN_TYPE
        ZOLTAN_OBJ_SIZE_FN_TYPE
        ZOLTAN_PACK_OBJ_FN_TYPE
        ZOLTAN_UNPACK_OBJ_FN_TYPE
        ZOLTAN_NUM_COARSE_OBJ_FN_TYPE
        ZOLTAN_COARSE_OBJ_LIST_FN_TYPE
        ZOLTAN_FIRST_COARSE_OBJ_FN_TYPE
        ZOLTAN_NEXT_COARSE_OBJ_FN_TYPE
        ZOLTAN_NUM_CHILD_FN_TYPE
        ZOLTAN_CHILD_LIST_FN_TYPE
        ZOLTAN_CHILD_WEIGHT_FN_TYPE
        ZOLTAN_OBJ_SIZE_MULTI_FN_TYPE
        ZOLTAN_PACK_OBJ_MULTI_FN_TYPE
        ZOLTAN_UNPACK_OBJ_MULTI_FN_TYPE
        ZOLTAN_PART_FN_TYPE
        ZOLTAN_PART_MULTI_FN_TYPE
        ZOLTAN_PROC_NAME_FN_TYPE
        ZOLTAN_HG_SIZE_CS_FN_TYPE
        ZOLTAN_HG_CS_FN_TYPE
        ZOLTAN_HG_SIZE_EDGE_WTS_FN_TYPE
        ZOLTAN_HG_EDGE_WTS_FN_TYPE
        ZOLTAN_NUM_FIXED_OBJ_FN_TYPE
        ZOLTAN_FIXED_OBJ_LIST_FN_TYPE
        ZOLTAN_HIER_NUM_LEVELS_FN_TYPE
        ZOLTAN_HIER_PART_FN_TYPE
        ZOLTAN_HIER_METHOD_FN_TYPE
        ZOLTAN_MAX_FN_TYPES

    ctypedef Zoltan_Fn_Type ZOLTAN_FN_TYPE

    # ************************************************************************
    # * Enumerated type used to indicate what type of refinement was used when
    # * building a refinement tree.
    # */
    enum Zoltan_Ref_Type:
        ZOLTAN_OTHER_REF
        ZOLTAN_IN_ORDER
        ZOLTAN_TRI_BISECT
        ZOLTAN_QUAD_QUAD
        ZOLTAN_HEX3D_OCT

    ctypedef Zoltan_Ref_Type ZOLTAN_REF_TYPE

    # /***********************************************************************
    # /***********************************************************************
    # /*******  Functions to set-up Zoltan load-balancing data structure  ****
    # /***********************************************************************
    # /***********************************************************************

    # /*
    #  *  Function to initialize values needed in load balancing tools, and
    #  *  returns which version of the library this is. If the application
    #  *  uses MPI, call this function after calling MPI_Init. If the
    #  *  application does not use MPI, this function calls MPI_Init for
    #  *  use by Zoltan. This function returns the version of
    #  *  the Zoltan library.
    #  *  Input:
    #  *    argc                --  Argument count from main()
    #  *    argv                --  Argument list from main()
    #  *  Output:
    #  *    ver                 --  Version of Zoltan library
    #  *  Returned value:       --  Error code
    #  */
    extern int Zoltan_Initialize(int, char**, float* ver)

    # Zoltan structure
    struct Zoltan_Struct:
        pass

    # function to create a zoltan structure
    extern Zoltan_Struct* Zoltan_Create(MPI_Comm)

    # /*****************************************************************************/
    # /*
    # *  Function to free the space associated with a Zoltan structure.
    # *  The input pointer is set to NULL when the routine returns.
    # *  Input/Output:
    # *    zz                  --  Pointer to a Zoltan structure.
    # */

    extern void Zoltan_Destroy(
        Zoltan_Struct **zz
        )

    # /*****************************************************************************/
    # /*
    #  *  Function to change a parameter value within Zoltan.
    #  *  Default values will be used for all parameters not explicitly altered
    #  *  by a call to this routine.
    #  *
    #  *  Input
    #  *    zz                  --  The Zoltan structure to which this
    #  *                            parameter alteration applies.
    #  *    name                --  The name of the parameter to have its
    #  *                            value changed.
    #  *    val                 --  The new value of the parameter.
    #  *
    #  *  Returned value:       --  Error code
    #  */
    extern int Zoltan_Set_Param(
        Zoltan_Struct *zz, char *name, char *val )

    #     /*****************************************************************************/
    # /*
    #  *  Function to return, for the calling processor, the number of objects
    #  *  located in that processor's memory.
    #  *  Input:
    #  *    data                --  pointer to user defined data structure
    #  *  Output:
    #  *    ierr                --  error code
    #  *  Returned value:       --  the number of local objects.
    #  */

    ctypedef int ZOLTAN_NUM_OBJ_FN(
        void *data,
        int *ierr
        ) except? -1

    extern int Zoltan_Set_Num_Obj_Fn(
        Zoltan_Struct *zz, ZOLTAN_NUM_OBJ_FN *fn_ptr, void *data_ptr)

    # /*****************************************************************************/
    # /*
    #  *  Function to return a list of all local objects on a processor.
    #  *  Input:
    #  *    data                --  pointer to user defined data structure
    #  *    num_gid_entries     --  number of array entries of type ZOLTAN_ID_TYPE
    #  *                            in a global ID
    #  *    num_lid_entries     --  number of array entries of type ZOLTAN_ID_TYPE
    #  *                            in a local ID
    #  *    wdim                --  dimension of object weights, or 0 if
    #  *                            object weights are not sought.
    #  *  Output:
    #  *    global_ids          --  array of Global IDs of all objects on the
    #  *                            processor.
    #  *    local_ids           --  array of Local IDs of all objects on the
    #  *                            processor.
    #  *    objwgts             --  objwgts[i*wdim:(i+1)*wdim-1] correponds
    #  *                            to the weight of object i
    #  *    ierr                --  error code
    #  */

    ctypedef void ZOLTAN_OBJ_LIST_FN(
        void *data,
        int num_gid_entries,
        int num_lid_entries,
        ZOLTAN_ID_PTR global_ids,
        ZOLTAN_ID_PTR local_ids,
        int wdim,
        float *objwgts,
        int *ierr
        )

    extern int Zoltan_Set_Obj_List_Fn(
        Zoltan_Struct *zz,
        ZOLTAN_OBJ_LIST_FN *fn_ptr,
        void *data_ptr
        )

    # /*****************************************************************************/
    # /*
    #  *  Function to return
    #  *  the number of geometry fields per object (e.g., the number of values
    #  *  used to express the coordinates of the object).
    #  *  Input:
    #  *    data                --  pointer to user defined data structure
    #  *  Output:
    #  *    ierr                --  error code
    #  *  Returned value:       --  the number of geometry fields.
    #  */

    # A ZOLTAN_NUM_GEOM_FN query function returns the number of values
    # needed to express the geometry of an object. For example, for a
    # two-dimensional mesh-based application, (x,y) coordinates are needed
    # to describe an object's geometry; thus the ZOLTAN_NUM_GEOM_FN query
    # function should return the value of two. For a similar
    # three-dimensional application, the return value should be three.

    ctypedef int ZOLTAN_NUM_GEOM_FN(
        void *data,
        int *ierr
        )  except? -1

    extern int Zoltan_Set_Num_Geom_Fn(
        Zoltan_Struct *zz,
        ZOLTAN_NUM_GEOM_FN *fn_ptr,
        void *data_ptr
        )

    # /*****************************************************************************/

    # A ZOLTAN_GEOM_MULTI FN query function returns a vector of geometry
    # values for a list of given objects. The geometry vector is
    # allocated by Zoltan to be of size num_obj * num_dim; its format is
    # described below

    # /*
    # *  Function to return the geometry information (e.g., coordinates) for
    # *  all objects.
    # *  Input:
    #     *    data                --  pointer to user defined data structure
    #     *    num_gid_entries     --  number of array entries of type ZOLTAN_ID_TYPE
    #     *                            in a global ID
    #     *    num_lid_entries     --  number of array entries of type ZOLTAN_ID_TYPE
    #     *                            in a local ID
    #     *    num_obj             --  number of objects whose coordinates are needed.
    #     *    global_id           --  array of Global IDs for the objects
    #     *    local_id            --  array of Local IDs for the objects;
    #     *                            NULL if num_lid_entries == 0.
    #     *    num_dim             --  dimension of coordinates for each object.
    #     *  Output:
    #     *    geom_vec            --  the geometry info for the objects;
    #     *                            (e.g., coordinates)
    #     *                            If num_dim == n, geom_vec[i*n+j] is the
    #     *                            jth coordinate for object i.
    #     *    ierr                --  error code
    #     */

    ctypedef void ZOLTAN_GEOM_MULTI_FN(
        void *data,
        int num_gid_entries,
        int num_lid_entries,
        int num_obj,
        ZOLTAN_ID_PTR global_id,
        ZOLTAN_ID_PTR local_id,
        int num_dim,
        double *geom_vec,
        int *ierr
        )

    extern int Zoltan_Set_Geom_Multi_Fn(
        Zoltan_Struct *zz,
        ZOLTAN_GEOM_MULTI_FN *fn_ptr,
        void *data_ptr
        )

    # /*****************************************************************************/
    # /*
    #  *  Function to invoke the partitioner.
    #  *
    #  *  Input:
    #  *    zz                  --  The Zoltan structure returned by Zoltan_Create.
    #  *  Output:
    #  *    changes             --  This value tells whether the new
    #  *                            decomposition computed by Zoltan differs
    #  *                            from the one given as input to Zoltan.
    #  *                            It can be either a one or a zero:
    #  *                            zero - No changes to the decomposition
    #  *                                   were made by the partitioning
    #  *                                   algorithm; migration isn't needed.
    #  *                            one  - A new decomposition is suggested
    #  *                                   by the partitioner; migration
    #  *                                   is needed to establish the new
    #  *                                   decomposition.
    #  *    num_gid_entries     --  number of entries of type ZOLTAN_ID_TYPE
    #  *                            in a global ID
    #  *    num_lid_entries     --  number of entries of type ZOLTAN_ID_TYPE
    #  *                            in a local ID
    #  *    num_import          --  The number of non-local objects in the
    #  *                            processor's new decomposition (i.e.,
    #  *                            number of objects to be imported).
    #  *    import_global_ids   --  Pointer to array of Global IDs for the
    #  *                            objects to be imported.
    #  *    import_local_ids    --  Pointer to array of Local IDs for the
    #  *                            objects to be imported (local to the
    #  *                            exporting processor).
    #  *    import_procs        --  Pointer to array of Processor IDs for the
    #  *                            objects to be imported (processor IDs of
    #  *                            source processor).
    #  *    import_to_part      --  Pointer to array of partition numbers to
    #  *                            which the imported objects should be assigned.
    #  *    num_export          --  The number of local objects that must be
    #  *                            exported from the processor to establish
    #  *                            the new decomposition.
    #  *    export_global_ids   --  Pointer to array of Global IDs for the
    #  *                            objects to be exported from the current
    #  *                            processor.
    #  *    export_local_ids    --  Pointer to array of Local IDs for the
    #  *                            objects to be exported (local to the
    #  *                            current processor).
    #  *    export_procs        --  Pointer to array of Processor IDs for the
    #  *                            objects to be exported (processor IDs of
    #  *                            destination processors).
    #  *    export_to_part      --  Pointer to array of partition numbers to
    #  *                            which the exported objects should be assigned.
    #  *  Returned value:       --  Error code
    #  */
    extern int Zoltan_LB_Partition(
        Zoltan_Struct *zz,
        int *changes,
        int *num_gid_entries,
        int *num_lid_entries,
        int *num_import,
        ZOLTAN_ID_PTR *import_global_ids,
        ZOLTAN_ID_PTR *import_local_ids,
        int **import_procs,
        int **import_to_part,
        int *num_export,
        ZOLTAN_ID_PTR *export_global_ids,
        ZOLTAN_ID_PTR *export_local_ids,
        int **export_procs,
        int **export_to_part
        )

    # /*****************************************************************************/
    # /*
    #  *  Function to invoke the load-balancer.
    #  *  Appropriate only when the number of requested partitions is equal to the
    #  *  number of processors.
    #  *
    #  *  Input and output:
    #  *    Arguments are analogous to Zoltan_LB_Partition.  Arrays import_to_part
    #  *    and export_to_part are not included, as Zoltan_LB_Balance assumes
    #  *    partitions and processors are equivalent.
    #  *  Returned value:       --  Error code
    #  */
    extern int Zoltan_LB_Balance(
        Zoltan_Struct *zz,
        int *changes,
        int *num_gid_entries,
        int *num_lid_entries,
        int *num_import,
        ZOLTAN_ID_PTR *import_global_ids,
        ZOLTAN_ID_PTR *import_local_ids,
        int **import_procs,
        int *num_export,
        ZOLTAN_ID_PTR *export_global_ids,
        ZOLTAN_ID_PTR *export_local_ids,
        int **export_procs
        )

    # /*****************************************************************************/
    # /*
    #  *  Routine to free the data arrays returned by Zoltan_Balance.  The arrays
    #  *  are freed and the pointers are set to NULL.
    #  *
    #  *  Input:
    #  *    import_global_ids   --  Pointer to array of global IDs for
    #  *                            imported objects.
    #  *    import_local_ids    --  Pointer to array of local IDs for
    #  *                            imported objects.
    #  *    import_procs        --  Pointer to array of processor IDs of
    #  *                            imported objects.
    #  *    export_global_ids   --  Pointer to array of global IDs for
    #  *                            exported objects.
    #  *    export_local_ids    --  Pointer to array of local IDs for
    #  *                            exported objects.
    #  *    export_procs        --  Pointer to array of destination processor
    #  *                            IDs of exported objects.
    #  *  Returned value:       --  Error code
    #  */

    extern int Zoltan_LB_Free_Data(
        ZOLTAN_ID_PTR *import_global_ids,
        ZOLTAN_ID_PTR *import_local_ids,
        int **import_procs,
        ZOLTAN_ID_PTR *export_global_ids,
        ZOLTAN_ID_PTR *export_local_ids,
        int **export_procs
        )

    # /*****************************************************************************/
    # /*
    #  *  Function to return the bounding box of a partition generated by RCB.
    #  *  Input:
    #  *    zz                  --  The Zoltan structure returned by Zoltan_Create.
    #  *    part                --  The partition number whose bounding box is to
    #  *                            be returned.
    #  *  Output:
    #  *    ndim                --  Number of dimensions in the bounding box.
    #  *    xmin                --  lower x extent of box
    #  *    ymin                --  lower y extent of box
    #  *    zmin                --  lower z extent of box
    #  *    xmax                --  upper x extent of box
    #  *    ymax                --  upper y extent of box
    #  *    zmax                --  upper z extent of box
    #  *  Returned value:       --  Error code
    #  */

    int Zoltan_RCB_Box(
        Zoltan_Struct *zz,
        int     part,
        int    *ndim,
        double *xmin,
        double *ymin,
        double *zmin,
        double *xmax,
        double *ymax,
        double *zmax
        )

    # /*****************************************************************************/
    # /*
    #  * Routine to determine which partitions and processors
    #  * a bounding box intersects.
    #  * Note that this only works of the current partition was produced via a
    #  * geometric algorithm - currently RCB, RIB, HSFC.
    #  *
    #  * Input:
    #  *   zz                   -- pointer to Zoltan structure
    #  *   xmin, ymin, zmin     -- lower left corner of bounding box
    #  *   xmax, ymax, zmax     -- upper right corner of bounding box
    #  *
    #  * Output:
    #  *   procs                -- list of processors that box intersects.
    #  *                           Note: application is
    #  *                               responsible for ensuring sufficient memory.
    #  *   numprocs             -- number of processors box intersects
    #  *   parts                -- list of partitions that box intersects.
    #  *                           Note: application is
    #  *                               responsible for ensuring sufficient memory.
    #  *   numparts             -- number of partitions box intersects (may differ
    #  *                           from numprocs).
    #  *
    #  * Returned value:       --  Error code
    #  */

    extern int Zoltan_LB_Box_PP_Assign(
        Zoltan_Struct *zz,
        double xmin,
        double ymin,
        double zmin,
        double xmax,
        double ymax,
        double zmax,
        int *procs,
        int *numprocs,
        int *parts,
        int *numparts
        )

    # /*****************************************************************************/
    # /*
    #  * Routine to determine which processor and partition a new point should be
    #  * assigned to.
    #  * Note that this only works of the current partition was produced via a
    #  * geometric algorithm - currently RCB, RIB, HSFC.
    #  *
    #  * Input:
    #  *   zz                   -- pointer to Zoltan structure
    #  *   coords               -- vector of coordinates of new point
    #  *
    #  * Output:
    #  *   proc                 -- processor that point should be assigned to
    #  *   part                 -- partition that point should be assigned to
    #  *
    #  *  Returned value:       --  Error code
    #  */

    extern int Zoltan_LB_Point_PP_Assign(
        Zoltan_Struct *zz,
        double *coords,
        int *proc,
        int *part
        )

    extern int Zoltan_LB_Point_Assign(
        Zoltan_Struct *zz,
        double *coords,
        int *proc
        )

    # /*****************************************************************************/
    # /*
    #  *  Routine to compute the inverse map:  Given, for each processor, a list
    #  *  of objects to be received by a processor, compute the list of objects
    #  *  that processor needs to send to other processors to complete a
    #  *  remapping.  Conversely, given a list of objects to be sent to other
    #  *  processors, compute the list of objects to be received.
    #  *
    #  *  Input:
    #  *    zz                  --  Zoltan structure for current
    #  *                            balance.
    #  *    num_input           --  Number of objects known to be
    #  *                            sent/received.
    #  *    input_global_ids    --  Array of global IDs for known objects.
    #  *    input_local_ids     --  Array of local IDs for known objects.
    #  *    input_procs         --  Array of IDs of processors to/from whom the
    #  *                            known objects will be sent/received.
    #  *    input_to_part       --  Array of partition numbers to
    #  *                            which the known objects should be assigned.
    #  *  Output:
    #  *    num_output          --  The number of objects will be received/sent.
    #  *    output_global_ids   --  Pointer to array of Global IDs for the
    #  *                            objects to be received/sent.
    #  *    output_local_ids    --  Pointer to array of Local IDs for the
    #  *                            objects to be received/sent.
    #  *    output_procs        --  Pointer to array of Processor IDs
    #  *                            from/to which the output_global_ids will be
    #  *                            received/sent.
    #  *    output_to_part      --  Pointer to array of partition numbers to
    #  *                            which the output_global_ids should be assigned.
    #  *  Returned value:       --  Error code
    #  */

    extern int Zoltan_Invert_Lists(
        Zoltan_Struct *zz,
        int num_input,
        ZOLTAN_ID_PTR input_global_ids,
        ZOLTAN_ID_PTR input_local_ids,
        int *input_procs,
        int *input_to_part,
        int *num_output,
        ZOLTAN_ID_PTR *output_global_ids,
        ZOLTAN_ID_PTR *output_local_ids,
        int **output_procs,
        int **output_to_part
        )

    # /*****************************************************************************/
    # /*
    # *  Wrapper around Zoltan_Invert_Lists, appropriate only when
    # *  number of partitions == number of processors (or when partition information
    #                                                  *  is not desired).
    # *
    # *  Input and Output:
    # *    Arguments are analogous to Zoltan_Invert_Lists.  Arrays import_to_part
    # *    and export_to_part are not included, as Zoltan_Compute_Destinations
    # *    assumes partitions and processors are equivalent.
    # *  Returned value:       --  Error code
    # */
    extern int Zoltan_Compute_Destinations(
        Zoltan_Struct *zz,
        int num_input,
        ZOLTAN_ID_PTR input_global_ids,
        ZOLTAN_ID_PTR input_local_ids,
        int *input_procs,
        int *num_output,
        ZOLTAN_ID_PTR *output_global_ids,
        ZOLTAN_ID_PTR *output_local_ids,
        int **output_procs
        )

    # /*****************************************************************************/
    # /*
    #  *  Routine to free the data arrays returned by Zoltan_LB_Partition,
    #  *  Zoltan_LB_Balance, Zoltan_Invert_Lists, and
    #  *  Zoltan_Compute_Destinations.  The arrays
    #  *  are freed and the pointers are set to NULL.
    #  *
    #  *  Input:
    #  *    global_ids   --  Pointer to array of global IDs
    #  *    local_ids    --  Pointer to array of local IDs
    #  *    procs        --  Pointer to array of processor IDs
    #  *    to_proc      --  Pointer to array of partition assignments
    #  *  Returned value:       --  Error code
    #  */
    extern int Zoltan_LB_Free_Part(
        ZOLTAN_ID_PTR *global_ids,
        ZOLTAN_ID_PTR *local_ids,
        int **procs,
        int **to_part
        )



# cython: language_level=3

cdef extern from "zoltan_types.h":

    # basic type used by all of Zoltan
    ctypedef unsigned int ZOLTAN_ID_TYPE

    # MPI data type
    cdef unsigned int ZOLTAN_ID_MPI_TYPE

    # pointer to the basic type
    ctypedef ZOLTAN_ID_TYPE* ZOLTAN_ID_PTR

    # /*****************************************************************************/
    # /*
    #  * Error codes for Zoltan library
    #  *   ZOLTAN_OK     - no errors
    #  *   ZOLTAN_WARN   - some warning occurred in Zoltan library;
    #  *                   application should be able to continue running
    #  *   ZOLTAN_FATAL  - a fatal error occurred
    #  *   ZOLTAN_MEMERR - memory allocation failed; with this error, it could be
    #  *                   possible to try a different, more memory-friendly,
    #  *                   algorithm.
    #  */
    # /*****************************************************************************/
    cdef int ZOLTAN_OK
    cdef int ZOLTAN_WARN
    cdef int ZOLTAN_FATAL
    cdef int ZOLTAN_MEMERR

    # /*****************************************************************************/
    # /* Hypergraph query function types
    #  */
    # /*****************************************************************************/
    cdef int _ZOLTAN_COMPRESSED_EDGE
    cdef int _ZOLTAN_COMPRESSED_VERTEX



__version__ = '1.1'


def get_include():
    """Return the directory in the package that contains header files."""
    from os.path import dirname, abspath, join
    return abspath(join(dirname(__file__), '../'))


_has_mpi = None


def has_mpi():
    """Return True if mpi4py is available.
    """
    global _has_mpi
    if _has_mpi is None:
        _has_mpi = True
        try:
            import mpi4py  # noqa: 401
        except ImportError:
            _has_mpi = False
        else:
            mpi4py.rc.initialize = False
            mpi4py.rc.finalize = True
    return _has_mpi


# Call this to disable mpi4py from initializing MPI on import.
has_mpi()



"""Running script for Zoltan"""

import unittest
from pytest import mark, importorskip

from pyzoltan.tools import run_parallel_script

path = run_parallel_script.get_directory(__file__)


class PyZoltanTests(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        importorskip("mpi4py.MPI")
        importorskip("pyzoltan.core.zoltan")

    def test_zoltan_geometric_partitioner(self):
        run_parallel_script.run(
            filename='geometric_partitioner.py', nprocs=4, path=path
        )

    def test_zoltan_partition(self):
        run_parallel_script.run(
            filename='3d_partition.py', nprocs=4, timeout=90.0, path=path
        )

    def test_zoltan_zcomm(self):
        run_parallel_script.run(
            filename='zcomm.py', nprocs=4, path=path
        )


if __name__ == '__main__':
    unittest.main()



# cython: language_level=3

cdef extern from "Zoltan_config.h":

    # package string
    #cdef str PACKAGE_STRING

    # package version
    #cdef str PACKAGE_VERSION

    # ZOLTAN_ID_TYPE
    cdef bint UNSIGNED_INT_GLOBAL_IDS



# cython: language_level=3, embedsignature=True
# distutils: language=c++

"""Wrapper for the Zoltan Distributed Directory and routines"""
cimport mpi4py.MPI as mpi
if MPI4PY_V2:
   from mpi4py.libmpi cimport MPI_Comm
else:
   from mpi4py.mpi_c cimport MPI_Comm

from cyarray.carray cimport UIntArray, IntArray

from pyzoltan.czoltan.czoltan_dd cimport *
from pyzoltan.czoltan.czoltan_types cimport ZOLTAN_ID_TYPE, ZOLTAN_ID_PTR


cdef class Zoltan_DD:
    # Pointer to the Zoltan DD
    cdef Zoltan_DD_Directory* dd

    # MPI communicator
    cdef MPI_Comm comm



"""Tests for the Zoltan unstructured communication package"""
import mpi4py.MPI as mpi
import numpy as np
from numpy import random

# import the unstructured communication package
from pyzoltan.core import zoltan_comm
from pyzoltan.core import zoltan

# MPI comm, rank and size
comm = mpi.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

# Initialize the Zoltan library
zoltan_version = zoltan.Zoltan_Initialize()

# each processor creates some random data
numObjectsTotal = 1 << 10

x = random.random(numObjectsTotal)
gids = np.array(np.arange(size * numObjectsTotal)
                )[rank * numObjectsTotal:(rank + 1) * numObjectsTotal]
gids = gids.astype(np.uint32)

# arbitrarily assign some objects to be sent to some other processor
nsend = np.int32(random.random_integers(low=1, high=5))
object_ids = random.random_integers(low=0, high=numObjectsTotal, size=nsend)
proclist = random.random_integers(
    low=0, high=size - 1, size=nsend
).astype(np.int32)

my_indices = np.where(proclist == rank)[0]
proclist[my_indices] = (rank + 1) % size

# create the ZComm object
tag = np.int32(0)
zcomm = zoltan_comm.ZComm(comm, tag=tag, nsend=nsend, proclist=proclist)

# the data to send and receive
senddata = x[object_ids]
recvdata = np.ones(zcomm.nreturn)

# use zoltan to exchange doubles
print("Proc %d, Sending %s to %s" % (rank, senddata, proclist))
zcomm.Comm_Do(senddata, recvdata)
print("Proc %d, Received %s" % (rank, recvdata))

# use zoltan to exchange unsigned ints
senddata = gids[object_ids]
recvdata = np.ones(zcomm.nreturn, dtype=np.uint32)
zcomm.set_nbytes(recvdata.itemsize, recvdata.dtype)

print("Proc %d, Sending %s to %s" % (rank, senddata, proclist))
zcomm.Comm_Do(senddata, recvdata)
print("Proc %d, Received %s" % (rank, recvdata))

# Test the Comm Reverse function
# modify the received data
recvdata[:] = rank

updated_info = np.zeros(zcomm.nsend, dtype=senddata.dtype)
print('Proc %d, sending updated data %s' % (rank, recvdata))
zcomm.Comm_Do_Reverse(recvdata, updated_info)
print('Proc %d, received updated data %s' % (rank, updated_info))



# cython: language_level=3, embedsignature=True
# distutils: language=c++

"""Cython wrapper for the Zoltan Distributed Directory"""

if MPI4PY_V2:
   from mpi4py.libmpi cimport MPI_Comm
else:
   from mpi4py.mpi_c cimport MPI_Comm

from czoltan_types cimport ZOLTAN_ID_TYPE, ZOLTAN_ID_PTR

cdef extern from "zoltan_dd.h":

    struct Zoltan_DD_Struct:
        pass

    ctypedef Zoltan_DD_Struct Zoltan_DD_Directory


    #/*********** Distributed Directory Function Prototypes ************/
    int Zoltan_DD_Create( Zoltan_DD_Directory** dd, MPI_Comm comm,
                          int num_gid, int num_lid, int user_length,
                          int table_length, int debug_level)

    void Zoltan_DD_Destroy( Zoltan_DD_Directory** dd)

    int Zoltan_DD_Update (Zoltan_DD_Directory *dd, ZOLTAN_ID_PTR gid,
                          ZOLTAN_ID_PTR lid, char *user, int *partition, int count)

    int Zoltan_DD_Find (Zoltan_DD_Directory *dd, ZOLTAN_ID_PTR gid,
                        ZOLTAN_ID_PTR lid, char *data, int *partition, int count,
                        int *owner)

    int Zoltan_DD_GetLocalKeys( Zoltan_DD_Directory* dd, ZOLTAN_ID_PTR* gid,
                                int* size )

    int Zoltan_DD_Remove (Zoltan_DD_Directory *dd, ZOLTAN_ID_PTR gid,
                          int count)

    void Zoltan_DD_Stats (Zoltan_DD_Directory *dd)

    int Zoltan_DD_Print (Zoltan_DD_Directory *dd)



# cython: language_level=3, embedsignature=True
# distutils: language=c++

"""Cython wrapper for the Zoltan unstructured communication package"""

if MPI4PY_V2:
   from mpi4py.libmpi cimport MPI_Comm
else:
   from mpi4py.mpi_c cimport MPI_Comm


cdef extern from "zoltan_comm.h":

    struct Zoltan_Comm_Obj:
        pass

    ctypedef Zoltan_Comm_Obj ZOLTAN_COMM_OBJ

    #/* function prototypes */

    int Zoltan_Comm_Create(ZOLTAN_COMM_OBJ**, int, int*, MPI_Comm, int, int*)

    int Zoltan_Comm_Destroy(ZOLTAN_COMM_OBJ**)

    int Zoltan_Comm_Do     (ZOLTAN_COMM_OBJ*, int, char*, int, char*)
    int Zoltan_Comm_Do_Post(ZOLTAN_COMM_OBJ*, int, char*, int, char*)
    int Zoltan_Comm_Do_Wait(ZOLTAN_COMM_OBJ*, int, char*, int, char*)
    int Zoltan_Comm_Do_AlltoAll(ZOLTAN_COMM_OBJ*, char*, int, char*)

    int Zoltan_Comm_Do_Reverse     (ZOLTAN_COMM_OBJ*, int, char*, int, int*, char*)
    int Zoltan_Comm_Do_Reverse_Post(ZOLTAN_COMM_OBJ*, int, char*, int, int*, char*)
    int Zoltan_Comm_Do_Reverse_Wait(ZOLTAN_COMM_OBJ*, int, char*, int, int*, char*)

    int Zoltan_Comm_Info(ZOLTAN_COMM_OBJ*, int*, int*, int*, int*, int*, int*, int*,
                         int*, int*, int*, int*, int*, int*)

    int Zoltan_Comm_Invert_Plan(ZOLTAN_COMM_OBJ**)

    int Zoltan_Comm_Resize(ZOLTAN_COMM_OBJ*, int*, int, int*)



'''Setup script for PyZoltan.

You can use some environment variables to control the build. Setting CC/CXX
will let you choose a custom compiler (these can also be set in the config file
discussed below). You can also export ZOLTAN or ZOLTAN_INCLUDE/ZOLTAN_LIBRARY.

These are more configuration file options that trump everything. The file is in
~/.compyle/config.py.  The options are:

# MPI options: handy on clusters like a Cray.
MPI_CFLAGS = ['...']  # must be a list.
MPI_LINK = ['...']

# Zoltan options
USE_TRILINOS = 1  # When set to anything, use "-ltrilinos_zoltan".
ZOLTAN = '/path/to_zoltan'  # looks inside this for $ZOLTAN/include/, lib/

# Not needed if using ZOLTAN
ZOLTAN_INCLUDE = 'path/include'  # path to zoltan.h
ZOLTAN_LIBRARY = 'path/lib'  # path to libzoltan.a

'''

import os
from subprocess import check_output
import sys


# This is taken from compyle.ext_module.
def get_config_file_opts():
    '''A global configuration file is used to configure build options
    for compyle and other packages.  This is located in:

    ~/.compyle/config.py

    The file can contain arbitrary Python that is exec'd. The variables defined
    here specify the compile and link args. For example, one may set:

    OMP_CFLAGS = ['-fopenmp']
    OMP_LINK = ['-fopenmp']

    Will use these instead of the defaults that are automatically determined.
    These must be lists.

    '''
    fname = os.path.expanduser(os.path.join('~', '.compyle', 'config.py'))
    opts = {}
    if os.path.exists(fname):
        print('Reading configuration options from %s.' % fname)
        with open(fname) as fp:
            exec(compile(fp.read(), fname, 'exec'), opts)
        opts.pop('__builtins__', None)
    return opts


# NOTE: the configuration options in the file trump everything else!
# These are options from the .compyle/config.py
CONFIG_OPTS = get_config_file_opts()


if len(os.environ.get('COVERAGE', '')) > 0:
    MACROS = [("CYTHON_TRACE", "1"), ("CYTHON_TRACE_NOGIL", "1")]
    COMPILER_DIRECTIVES = {"linetrace": True}
    print("-" * 80)
    print("Enabling linetracing for cython and setting CYTHON_TRACE = 1")
    print("-" * 80)
else:
    MACROS = []
    COMPILER_DIRECTIVES = {}

MODE = 'normal'
if len(sys.argv) >= 2 and \
   ('--help' in sys.argv[1:] or
    sys.argv[1] in ('--help-commands', 'egg_info', '--version',
                    'clean', 'sdist')):
    MODE = 'info'

HAVE_MPI = True
USE_ZOLTAN = True
try:
    import mpi4py
except ImportError:
    HAVE_MPI = False

compiler = 'gcc'
# compiler = 'mpich'
# compiler = 'intel'
if compiler == 'intel':
    extra_compile_args = ['-O3']
else:
    extra_compile_args = []


def get_deps(*args):
    """Given a list of basenames, this checks if a .pyx or .pxd exists
    and returns the list.
    """
    result = []
    for basename in args:
        for ext in ('.pyx', '.pxd'):
            f = basename + ext
            if os.path.exists(f):
                result.append(f)
    return result


def get_zoltan_directory(varname):
    global USE_ZOLTAN
    if varname in CONFIG_OPTS:
        return os.path.expanduser(CONFIG_OPTS[varname])

    d = os.environ.get(varname, '')
    if len(d) == 0:
        USE_ZOLTAN = False
        return ''
    else:
        USE_ZOLTAN = True
    if not os.path.exists(d):
        print("*" * 80)
        print("%s incorrectly set to %s, not using ZOLTAN!" % (varname, d))
        print("*" * 80)
        USE_ZOLTAN = False
        return ''
    return d


def get_mpi_flags():
    """Returns mpi_inc_dirs, mpi_compile_args, mpi_link_args.
    """
    global HAVE_MPI
    mpi_inc_dirs = []
    mpi_compile_args = []
    mpi_link_args = []
    if not HAVE_MPI:
        return mpi_inc_dirs, mpi_compile_args, mpi_link_args
    elif 'MPI_CFLAGS' in CONFIG_OPTS:
        mpi_compile_args = CONFIG_OPTS['MPI_CFLAGS']
        mpi_link_args = CONFIG_OPTS['MPI_LINK']
        mpi_inc_dirs.append(mpi4py.get_include())
    else:
        try:
            mpic = 'mpic++'
            if compiler == 'intel':
                link_args = check_output(
                    [mpic, '-cc=icc', '-link_info'],
                    universal_newlines=True
                ).strip()
                link_args = link_args[3:]
                compile_args = check_output(
                    [mpic, '-cc=icc', '-compile_info'],
                    universal_newlines=True
                ).strip()
                compile_args = compile_args[3:]
            elif compiler == 'mpich':
                link_args = check_output(
                    [mpic, '-link-info'],
                    universal_newlines=True
                ).strip()
                link_args = " ".join(link_args.split()[1:])
                compile_args = check_output(
                    [mpic, '-compile_info'],
                    universal_newlines=True
                ).strip()
                compile_args = " ".join(compile_args.split()[1:])
            else:
                link_args = check_output(
                    [mpic, '--showme:link'],
                    universal_newlines=True
                ).strip()
                compile_args = check_output(
                    [mpic, '--showme:compile'],
                    universal_newlines=True
                ).strip()
        except:  # noqa: E722
            print('-' * 80)
            print("Unable to run mpic++ correctly, skipping parallel build")
            print('-' * 80)
            HAVE_MPI = False
        else:
            mpi_link_args.extend(link_args.split())
            mpi_compile_args.extend(compile_args.split())
            mpi_inc_dirs.append(mpi4py.get_include())

    return mpi_inc_dirs, mpi_compile_args, mpi_link_args


def get_zoltan_args():
    """Returns zoltan_include_dirs, zoltan_library_dirs
    """
    global HAVE_MPI, USE_ZOLTAN
    zoltan_include_dirs, zoltan_library_dirs = [], []
    if not HAVE_MPI:
        return zoltan_include_dirs, zoltan_library_dirs
    # First try with the environment variable 'ZOLTAN'
    zoltan_base = get_zoltan_directory('ZOLTAN')
    inc = lib = ''
    if len(zoltan_base) > 0:
        inc = os.path.join(zoltan_base, 'include')
        lib = os.path.join(zoltan_base, 'lib')
        if not os.path.exists(inc) or not os.path.exists(lib):
            inc = lib = ''

    # try with the older ZOLTAN include directories
    if len(inc) == 0 or len(lib) == 0:
        inc = get_zoltan_directory('ZOLTAN_INCLUDE')
        lib = get_zoltan_directory('ZOLTAN_LIBRARY')

    if not USE_ZOLTAN:
        # Try with default in sys.prefix/{include,lib}, this is what is done
        # by any conda installs of zoltan.
        inc = os.path.join(sys.prefix, 'include')
        lib = os.path.join(sys.prefix, 'lib')
        if os.path.exists(os.path.join(inc, 'zoltan.h')):
            USE_ZOLTAN = True

    if (not USE_ZOLTAN):
        print("*" * 80)
        print("Zoltan Environment variable not set, not using ZOLTAN!")
        print("*" * 80)
        HAVE_MPI = False
    else:
        print('-' * 70)
        print("Using Zoltan from:\n%s\n%s" % (inc, lib))
        print('-' * 70)
        zoltan_include_dirs = [inc]
        zoltan_library_dirs = [lib]

        # PyZoltan includes
        zoltan_cython_include = [os.path.abspath('./pyzoltan/czoltan')]
        zoltan_include_dirs += zoltan_cython_include

    # Not sure we need this but doing so just to be safe.
    import cyarray
    cyarray_include_dirs = [os.path.abspath(os.path.dirname(cyarray.__file__))]
    zoltan_include_dirs += cyarray_include_dirs

    return zoltan_include_dirs, zoltan_library_dirs


def get_parallel_extensions():
    if not HAVE_MPI:
        return []

    if MODE == 'info':
        from distutils.core import Extension
        include_dirs = []
        mpi_inc_dirs, mpi_compile_args, mpi_link_args = [], [], []
        zoltan_include_dirs, zoltan_library_dirs = [], []
    else:
        from Cython.Distutils import Extension
        import numpy
        include_dirs = [numpy.get_include()]
        mpi_inc_dirs, mpi_compile_args, mpi_link_args = get_mpi_flags()
        zoltan_include_dirs, zoltan_library_dirs = get_zoltan_args()

    # We should check again here as HAVE_MPI may be set to False when we try to
    # get the MPI flags and are not successful.
    if not HAVE_MPI:
        return []

    MPI4PY_V2 = False if mpi4py.__version__.startswith('1.') else True
    cython_compile_time_env = {'MPI4PY_V2': MPI4PY_V2}

    zoltan_lib = 'zoltan'
    if 'USE_TRILINOS' in CONFIG_OPTS:
        zoltan_lib = 'trilinos_zoltan'
    elif os.environ.get('USE_TRILINOS', None) is not None:
        zoltan_lib = 'trilinos_zoltan'

    zoltan_modules = [
        Extension(
            name="pyzoltan.core.zoltan",
            sources=["pyzoltan/core/zoltan.pyx"],
            depends=get_deps(
                "pyzoltan/czoltan/czoltan",
                "pyzoltan/czoltan/czoltan_types",
            ),
            include_dirs=include_dirs + zoltan_include_dirs + mpi_inc_dirs,
            library_dirs=zoltan_library_dirs,
            libraries=[zoltan_lib, 'mpi'],
            extra_link_args=mpi_link_args,
            extra_compile_args=mpi_compile_args + extra_compile_args,
            cython_compile_time_env=cython_compile_time_env,
            define_macros=MACROS,
        ),

        Extension(
            name="pyzoltan.core.zoltan_dd",
            sources=["pyzoltan/core/zoltan_dd.pyx"],
            depends=get_deps(
                "pyzoltan/core/carray",
                "pyzoltan/czoltan/czoltan_dd",
                "pyzoltan/czoltan/czoltan_types"
            ),
            include_dirs=include_dirs + zoltan_include_dirs + mpi_inc_dirs,
            library_dirs=zoltan_library_dirs,
            libraries=[zoltan_lib, 'mpi'],
            extra_link_args=mpi_link_args,
            extra_compile_args=mpi_compile_args + extra_compile_args,
            cython_compile_time_env=cython_compile_time_env,
            define_macros=MACROS,
        ),

        Extension(
            name="pyzoltan.core.zoltan_comm",
            sources=["pyzoltan/core/zoltan_comm.pyx"],
            depends=get_deps(
                "pyzoltan/core/carray",
                "pyzoltan/czoltan/zoltan_comm"
            ),
            include_dirs=include_dirs + zoltan_include_dirs + mpi_inc_dirs,
            library_dirs=zoltan_library_dirs,
            libraries=[zoltan_lib, 'mpi'],
            extra_link_args=mpi_link_args,
            extra_compile_args=mpi_compile_args + extra_compile_args,
            cython_compile_time_env=cython_compile_time_env,
            define_macros=MACROS,
        ),
    ]

    return zoltan_modules


def _is_cythonize_default():
    import warnings
    result = True
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            # old_build_ext was introduced in Cython 0.25 and this is when
            # cythonize was made the default.
            from Cython.Distutils import old_build_ext  # noqa: F401
        except ImportError:
            result = False
    return result


def setup_package():
    from setuptools import find_packages, setup
    if MODE == 'info':
        cmdclass = {}
    else:
        from Cython.Distutils import build_ext
        cmdclass = {'build_ext': build_ext}

    # Extract the version information from pysph/__init__.py
    info = {}
    module = os.path.join('pyzoltan', '__init__.py')
    exec(compile(open(module).read(), module, 'exec'), info)

    # The requirements.
    tests_require = ["pytest>=3.0"]
    docs_require = ["sphinx"]

    ext_modules = get_parallel_extensions()
    if MODE != 'info' and _is_cythonize_default():
        # Cython >= 0.25 uses cythonize to compile the extensions. This
        # requires the compile_time_env to be set explicitly to work.
        compile_env = {}
        include_path = set()
        if HAVE_MPI:
            MPI4PY_V2 = False if mpi4py.__version__.startswith('1.') else True
            compile_env.update({'MPI4PY_V2': MPI4PY_V2})

        for mod in ext_modules:
            compile_env.update(mod.cython_compile_time_env or {})
            include_path.update(mod.include_dirs)
        from Cython.Build import cythonize
        ext_modules = cythonize(
            ext_modules, compile_time_env=compile_env,
            include_path=list(include_path),
            compiler_directives=COMPILER_DIRECTIVES,
        )
        if len(ext_modules) == 0:
            raise RuntimeError(
                'There are no extension modules, Nothing to do!'
            )

    setup(name='PyZoltan',
          version=info['__version__'],
          author='PySPH Developers',
          author_email='pysph-dev@googlegroups.com',
          description='Wrapper for the Zoltan data management library',
          long_description=open('README.rst').read(),
          url='http://github.com/pypr/pyzoltan',
          license="BSD",
          keywords="Cython Zoltan Dynamic load balancing",
          packages=find_packages(),
          package_data={
              '': ['*.pxd', '*.rst', '*.txt'],
              'pyzoltan.core.tests': ['mesh.txt'],
          },
          # exclude package data in installation.
          exclude_package_data={
              '': ['Makefile', '*.bat', '*.cfg', '*.rst', '*.sh', '*.yml'],
          },
          ext_modules=ext_modules,
          include_package_data=True,
          cmdclass=cmdclass,
          extras_require={
              "docs": docs_require,
              "tests": tests_require,
              "dev": docs_require + tests_require,
          },
          zip_safe=False,
          platforms=['Linux', 'Mac OS-X', 'Unix', 'Windows'],
          classifiers=[c.strip() for c in """\
            Development Status :: 5 - Production/Stable
            Environment :: Console
            Intended Audience :: Developers
            Intended Audience :: Science/Research
            License :: OSI Approved :: BSD License
            Natural Language :: English
            Operating System :: MacOS :: MacOS X
            Operating System :: Microsoft :: Windows
            Operating System :: POSIX
            Operating System :: Unix
            Programming Language :: Python
            Programming Language :: Python :: 3
            Topic :: Scientific/Engineering
            Topic :: Scientific/Engineering :: Physics
            Topic :: Software Development :: Libraries
            """.splitlines() if len(c.split()) > 0],
          )


if __name__ == '__main__':
    setup_package()



"""Clone of the Zoltan tests for the geometric partitioners (RCB, RIB
and HSFC)"""

from __future__ import print_function

from os.path import abspath, dirname, join
import sys

import mpi4py.MPI as mpi
comm = mpi.COMM_WORLD

from cyarray.api import UIntArray, DoubleArray
from pyzoltan.core import zoltan

MY_DIR = dirname(abspath(__file__))

ack_tag = 5
count_tag = 10
id_tag = 15
x_tag = 20
y_tag = 25
ack = 0

import numpy as np
from numpy import loadtxt

def read_input_file(fname=join(MY_DIR, 'mesh.txt')):
    rank = comm.Get_rank()
    size = comm.Get_size()

    if rank == 0:
        # read the data
        data_line = 0
        data = loadtxt( fname, dtype=int )
        numGlobalPoints = data[data_line][0]; data_line += 1

        if size > 1:
            nobj = numGlobalPoints // 2
            remaining = numGlobalPoints - nobj
        else:
            nobj = numGlobalPoints
            remaining = 0

        myGlobalIds = np.ones(shape=nobj, dtype=np.uint32)
        x = np.ones(shape=nobj, dtype=np.float64)
        y = np.ones(shape=nobj, dtype=np.float64)
        numMyPoints = nobj

        for i in range(nobj):
            myGlobalIds[i], x[i], y[i] = data[data_line]; data_line += 1

        gids = np.ones(shape=(nobj+1), dtype=np.uint32)
        xcoord = np.ones(shape=(nobj+1), dtype=np.float64)
        ycoord = np.ones(shape=(nobj+1), dtype=np.float64)

        for i in range(1,size):
            if (remaining > 1):
                nobj = remaining // 2
                remaining -= nobj

            elif (remaining == 1):
                nobj = 1
                remaining = 0

            else:
                nobj = 0

            if ( (i== size-1) and (remaining > 0) ):
                nobj += remaining

            if (nobj > 0):
                for j in range(nobj):
                    gids[j],xcoord[j],ycoord[j] = data[data_line];data_line += 1

            comm.send(obj=nobj, dest=i, tag=count_tag)
            ack = comm.recv(source=i, tag=ack_tag)

            if (nobj > 0):
                comm.send(gids[:nobj], dest=i, tag=id_tag)
                comm.send(xcoord[:nobj], dest=i, tag=x_tag)
                comm.send(ycoord[:nobj], dest=i, tag=y_tag)

        ack = 0
        for i in range(1, size):
            comm.send(ack, dest=i, tag=0)

    else:
        numMyPoints = comm.recv(source=0, tag=count_tag)
        ack = 0
        if (numMyPoints > 0):
            myGlobalIds = np.ones(shape=numMyPoints, dtype=np.uint32)
            x = np.ones(shape=numMyPoints, dtype=np.float64)
            y = np.ones(shape=numMyPoints, dtype=np.float64)

            comm.send(obj=ack, dest=0, tag=ack_tag)
            myGlobalIds[:] = comm.recv(source=0, tag=id_tag)
            x[:] = comm.recv(source=0, tag=x_tag)
            y[:] = comm.recv(source=0, tag=y_tag)

        elif ( numMyPoints == 0 ):
            comm.send(obj=ack, dest=0, tag=ack_tag)

        else:
            sys.exit(-1)

        ack = comm.recv( source=0, tag=0 )
        if ( ack < 0 ):
            sys.exit(-1)

    return numMyPoints, myGlobalIds, x, y

def show_mesh(rank, numPoints, gids, parts):

    partAssign = np.zeros(shape=25, dtype=np.int32)
    allPartAssign = np.zeros(shape=25, dtype=np.int32)
    i = j = part = 0

    for i in range(numPoints):
        partAssign[ gids[i]-1 ] = parts[i]

    comm.Reduce(sendbuf=partAssign,
                recvbuf=allPartAssign, op=mpi.MAX, root=0)

    if rank==0:

        for i in range(20, -1, -5):
            for j in range(5):
                part = allPartAssign[i+j]
                if j < 4:
                    print("%d----"%(part),)
                else:
                    print("%d"%(part))
            if i > 0:
                print("|     |     |     |     |")

        print()

# read the input file and distribute objects across processors
numMyPoints, myGlobalIds, x, y = read_input_file()
rank = comm.Get_rank()

parts = np.ones(shape=numMyPoints, dtype=np.int32)
for i in range(numMyPoints):
    parts[i] = rank

# now do the load balancing
_x = np.asarray(x); _y = np.asarray(y); _gid = np.asarray(myGlobalIds)
_z = np.zeros_like(x)

x = DoubleArray(numMyPoints); x.set_data(_x)
y = DoubleArray(numMyPoints); y.set_data(_y)
z = DoubleArray(numMyPoints); z.set_data(_z)
gid = UIntArray(numMyPoints); gid.set_data(_gid)

pz = zoltan.ZoltanGeometricPartitioner(dim=2, comm=comm, x=x, y=y, z=z, gid=gid)

# set the weights to 0 by default
weights = pz.weights.get_npy_array()
weights[:] = 0

pz.set_lb_method("RCB")
pz.Zoltan_Set_Param("DEBUG_LEVEL","0")
pz.Zoltan_LB_Balance()

if rank == 0:
    print("\nMesh partition before Zoltan\n")

comm.barrier()

show_mesh(rank, numMyPoints, myGlobalIds, parts)

for i in range(pz.numExport):
    parts[ pz.exportLocalids[i] ] = pz.exportProcs[i]

if rank == 0:
    print("Mesh partition assignment after calling Zoltan")

show_mesh(rank, numMyPoints, myGlobalIds, parts)



from __future__ import print_function
from os.path import abspath, dirname, join
from subprocess import Popen, PIPE
import sys
from threading import Timer


def get_directory(file):
    return dirname(abspath(file))


def kill_process(process):
    print('KILLING PROCESS ON TIMEOUT')
    process.kill()


def run(filename, args=None, nprocs=2, timeout=30.0, path=None,
        mpi_extra=None):
    """Run a python script with MPI or in serial (if nprocs=1).  Kill process
    if it takes longer than the specified timeout.

    Parameters:
    -----------
    filename - filename of python script to run under mpi.
    args - List of arguments to pass to script.
    nprocs - number of processes to run (1 => serial non-mpi run).
    timeout - time in seconds to wait for the script to finish running,
        else raise a RuntimeError exception.
    path - the path under which the script is located
        Defaults to the location of this file (__file__), not curdir.
    mpi_extra: List of additional arguments to mpiexec.

    """
    if args is None:
        args = []
    file_path = abspath(join(path, filename))
    extra = ['--oversubscribe'] if mpi_extra is None else mpi_extra
    cmd = extra + [sys.executable, file_path] + args
    if nprocs > 1:
        cmd = ['mpiexec', '-n', str(nprocs)] + cmd

    print('running test:', cmd)

    process = Popen(cmd, stdout=PIPE, stderr=PIPE)
    timer = Timer(timeout, kill_process, [process])
    timer.start()
    out, err = process.communicate()
    timer.cancel()
    retcode = process.returncode
    if retcode:
        msg = 'test ' + filename + ' failed with returncode ' + str(retcode)
        print(out.decode('utf-8'))
        print(err.decode('utf-8'))
        print('#'*80)
        print(msg)
        print('#'*80)
        raise RuntimeError(msg)
    return retcode, out, err



# cython: language_level=3, embedsignature=True
# distutils: language=c++

cimport mpi4py.MPI as mpi
if MPI4PY_V2:
   from mpi4py cimport libmpi as mpic
else:
   from mpi4py cimport mpi_c as mpic

# Zoltan imports
from pyzoltan.czoltan cimport czoltan
from pyzoltan.czoltan.czoltan cimport Zoltan_Struct

# Zoltan type imports
from pyzoltan.czoltan.czoltan_types cimport ZOLTAN_ID_PTR, ZOLTAN_ID_TYPE, \
     ZOLTAN_OK, ZOLTAN_WARN, ZOLTAN_FATAL, ZOLTAN_MEMERR

# NUMPY
import numpy as np
cimport numpy as np

# Carrays
from cyarray.carray cimport UIntArray, IntArray, LongArray, DoubleArray

# Compatibility for older MPI versions and later
# mpi4py releases (Ubuntu 14.04 is one such).
cdef extern from 'mpi-compat.h': pass

# Error checking for Zoltan
cdef _check_error(int ierr)

# Pointer to the Zoltan struct
cdef struct _Zoltan_Struct:
    czoltan.Zoltan_Struct* zz

cdef class PyZoltan:
    # problem dimensionsionality
    cdef public int dim

    # version number
    cdef public double version

    # mpi.Comm object and associated rank and size
    cdef public object comm
    cdef public int rank, size

    # Pointer to the Zoltan structure upon creation
    cdef _Zoltan_Struct _zstruct

    # string to store the current load balancing method
    cdef public str lb_method

    # Arrays returned by Zoltan
    cdef public UIntArray exportGlobalids
    cdef public UIntArray exportLocalids
    cdef public IntArray exportProcs

    cdef public UIntArray importGlobalids
    cdef public UIntArray importLocalids
    cdef public IntArray importProcs

    # the number of objects to import/export
    cdef public int numImport, numExport

    cdef public np.ndarray procs             # processors of range size
    cdef public np.ndarray parts             # partitions of range size

    # data array for the object weights
    cdef public DoubleArray weights

    # General Zoltan parameters (refer the user guide)
    cdef public str debug_level
    cdef public str obj_weight_dim
    cdef public str edge_weight_dim
    cdef public str return_lists

    ###############################################################
    # Member functions
    ###############################################################
    # after a load balance, copy the Zoltan allocated lists to local
    # numpy arrays. The Zoltan lists are subsequently deallocated
    cdef _set_zoltan_lists(
        self,
        int numExport,                          # number of objects to export
        ZOLTAN_ID_PTR _exportGlobal,            # global indices of export objects
        ZOLTAN_ID_PTR _exportLocal,             # local indices of export objects
        int* _exportProcs,                      # target processors to export
        int numImport,                          # number of objects to import
        ZOLTAN_ID_PTR _importGlobal,            # global indices of import objects
        ZOLTAN_ID_PTR _importLocal,             # local indices of import objects
        int* _importProcs                       # target processors to import
        )

    # Invert the export lists. Given a situation where every processor
    # knows which objects must be exported to remote processors, a
    # call to invert lists will return a list of objects that must be
    # imported from remote processors.
    cpdef Zoltan_Invert_Lists(self)

# User defined data for the RCB, RIB and HSFC methods
cdef struct CoordinateData:
    # flag for using weights
    bint use_weights

    # dimensionality of the problem
    int dim

    # number of local/global points
    int numGlobalPoints
    int numMyPoints

    # pointers to the object data
    ZOLTAN_ID_PTR myGlobalIDs
    double* obj_wts
    double* x
    double* y
    double *z

cdef class ZoltanGeometricPartitioner(PyZoltan):
    # data arrays for the coordinates
    cdef public DoubleArray x, y, z

    # data array for the global indices
    cdef public UIntArray gid

    # User defined structure to hold the coordinate data for the
    # Zoltan interface
    cdef CoordinateData _cdata

    # number of global and local objects
    cdef public int num_global_objects, num_local_objects

    # ZOLTAN parameters for Geometric partitioners
    cdef public str keep_cuts



"""General imports for PyZoltan"""

# CArrays
import cyarray.carray as carray
from cyarray.carray import UIntArray, IntArray, LongArray, \
    DoubleArray

# Main Zoltan load balancer
from pyzoltan.core.zoltan import get_zoltan_id_type_max
from pyzoltan.core.zoltan import PyZoltan, ZoltanGeometricPartitioner

# Zoltan unstructured comm
from pyzoltan.core.zoltan_comm import ZComm

# Zoltan distributed directory
from pyzoltan.core.zoltan_dd import Zoltan_DD
